@@ -17,16 +17,6 @@
 #define libGAP_GAP_CODE_H
 
 
-/****************************************************************************
-**
-
-*S  T_BODY  . . . . . . . . . . . . . . . . . . . . type of function body bag
-**
-**  'T_BODY' is the type of the function body bags.
-*/
-#define libGAP_T_BODY                  175
-
-
 /****************************************************************************
 **
 *T  Stat  . . . . . . . . . . . . . . . . . . . . . . . .  type of statements
@@ -36,7 +26,7 @@
 **  If 'Stat' is different  from 'Expr', then  a lot of things will  probably
 **  break.
 */
-#define libGAP_Stat            libGAP_UInt
+#define libGAP_Stat            libGAP_UInt8
 
 
 /****************************************************************************
@@ -163,7 +153,11 @@ extern  libGAP_Stat *          libGAP_PtrBody;
 
 #define libGAP_T_PROCCALL_OPTS         (libGAP_FIRST_STAT_TNUM+ 79)
 
-#define libGAP_LAST_STAT_TNUM          libGAP_T_PROCCALL_OPTS
+#define libGAP_T_ATOMIC               (libGAP_FIRST_STAT_TNUM+80)
+
+#define libGAP_LAST_STAT_TNUM          libGAP_T_ATOMIC
+
+#define libGAP_T_NO_STAT		(libGAP_Stat)(-1)
 
 
 /****************************************************************************
@@ -172,7 +166,7 @@ extern  libGAP_Stat *          libGAP_PtrBody;
 **
 **  'TNUM_STAT' returns the type of the statement <stat>.
 */
-#define libGAP_TNUM_STAT(stat) (libGAP_ADDR_STAT(stat)[-1] & 0xFF)
+#define libGAP_TNUM_STAT(stat) ((libGAP_Int)(libGAP_ADDR_STAT(stat)[-1] & 0xFF))
 
 
 /****************************************************************************
@@ -181,7 +175,43 @@ extern  libGAP_Stat *          libGAP_PtrBody;
 **
 **  'SIZE_STAT' returns the size of the statement <stat>.
 */
-#define libGAP_SIZE_STAT(stat) (libGAP_ADDR_STAT(stat)[-1] >> 8)
+#define libGAP_SIZE_STAT(stat) ((libGAP_Int)(libGAP_ADDR_STAT(stat)[-1] >> 8 & 0xFFFFFF))
+
+/****************************************************************************
+**
+*F  LINE_STAT(<stat>) . . . . . . . . . . . . . . line number of a statement
+**
+**  'LINE_STAT' returns the line number of the statement <stat>.
+*/
+#define libGAP_LINE_STAT(stat) ((libGAP_Int)(libGAP_ADDR_STAT(stat)[-1] >> 32 & 0xFFFF))
+
+/****************************************************************************
+**
+*F  FILENAMEID_STAT(<stat>) . . . . . . . . . . . . file name of a statement
+**
+**  'FILENAMEID_STAT' returns the file the statment <stat> was read from.
+**  This should be looked up in the FilenameCache variable
+*/
+#define libGAP_FILENAMEID_STAT(stat) ((libGAP_Int)(libGAP_ADDR_STAT(stat)[-1] >> 48 & 0x7FFF))
+
+/****************************************************************************
+**
+*F  FILENAME_STAT(<stat>) . . . . . . . . . . . . file name of a statement
+**
+**  'FILENAME_STAT' returns a gap string containing the file where the statment
+**  <stat> was read from.
+*/
+libGAP_Obj libGAP_FILENAME_STAT(libGAP_Stat stat);
+
+/****************************************************************************
+**
+*F  VISITED_STAT(<stat>) . . . . . . . . . . . . if statement has even been run
+**
+**  'VISITED_STAT' returns true if the statement has ever been executed
+**  while profiling is turned on.
+*/
+#define libGAP_VISITED_STAT(stat) (libGAP_ADDR_STAT(stat)[-1] >> 63 && 0x1)
+
 
 
 /****************************************************************************
@@ -191,7 +221,7 @@ extern  libGAP_Stat *          libGAP_PtrBody;
 **  'ADDR_STAT' returns   the  absolute address of the    memory block of the
 **  statement <stat>.
 */
-#define libGAP_ADDR_STAT(stat) ((libGAP_Stat*)(((char*)libGAP_PtrBody)+(stat)))
+#define libGAP_ADDR_STAT(stat) ((libGAP_Stat*)(((char*)libGAP_TLS(libGAP_PtrBody))+(stat)))
 
 
 /****************************************************************************
@@ -359,7 +389,12 @@ extern  libGAP_Stat *          libGAP_PtrBody;
 #define libGAP_T_FLOAT_EXPR_EAGER      (libGAP_FIRST_EXPR_TNUM+82)
 #define libGAP_T_FLOAT_EXPR_LAZY       (libGAP_FIRST_EXPR_TNUM+83)
 
-#define libGAP_LAST_EXPR_TNUM          libGAP_T_FLOAT_EXPR_LAZY
+#define libGAP_T_ELM2_LIST             (libGAP_FIRST_EXPR_TNUM+84)
+#define libGAP_T_ELMX_LIST             (libGAP_FIRST_EXPR_TNUM+85)
+#define libGAP_T_ASS2_LIST             (libGAP_FIRST_EXPR_TNUM+86)
+#define libGAP_T_ASSX_LIST             (libGAP_FIRST_EXPR_TNUM+87)
+
+#define libGAP_LAST_EXPR_TNUM          libGAP_T_ASSX_LIST
 
 
 /****************************************************************************
@@ -671,6 +706,51 @@ extern  void            libGAP_CodeForEndBody (
 
 extern  void            libGAP_CodeForEnd ( void );
 
+/****************************************************************************
+**
+*F  CodeAtomicBegin()  . . . . . . .  code atomic-statement, begin of statement
+*F  CodeAtomicBeginBody()  . . . . . . . . code atomic-statement, begin of body
+*F  CodeAtomicEndBody( <nr> )  . . . . . . . code atomic-statement, end of body
+*F  CodeAtomicEnd()  . . . . . . . . .  code atomic-statement, end of statement
+**
+**  'CodeAtomicBegin'  is an action to  code a atomic-statement.   It is called
+**  when the  reader encounters the 'atomic',  i.e., *before* the condition is
+**  read.
+**
+**  'CodeAtomicBeginBody'  is  an action   to code a  atomic-statement.   It is
+**  called when  the reader encounters  the beginning  of the statement body,
+**  i.e., *after* the condition is read.
+**
+**  'CodeAtomicEndBody' is an action to  code a atomic-statement.  It is called
+**  when the reader encounters  the end of  the statement body.  <nr> is  the
+**  number of statements in the body.
+**
+**  'CodeAtomicEnd' is an action to code a atomic-statement.  It is called when
+**  the reader encounters  the end  of the  statement, i.e., immediate  after
+**  'CodeAtomicEndBody'.
+**
+**  These functions are just placeholders for the future HPC-GAP code.
+*/
+
+void libGAP_CodeAtomicBegin ( void );
+
+void libGAP_CodeAtomicBeginBody ( libGAP_UInt nrexprs );
+
+void libGAP_CodeAtomicEndBody (
+    libGAP_UInt                nrstats );
+void libGAP_CodeAtomicEnd ( void );
+
+/****************************************************************************
+**
+*F  CodeQualifiedExprBegin()  . . . code readonly/readwrite expression start
+*F  CodeQualifiedExprEnd()  . . . . . code readonly/readwrite expression end
+**
+*/
+
+void libGAP_CodeQualifiedExprBegin(libGAP_UInt qual);
+
+void libGAP_CodeQualifiedExprEnd( void );
+
 
 /****************************************************************************
 **
@@ -1080,17 +1160,17 @@ extern  void            libGAP_CodeIsbGVar (
 *F  CodeAssListLevel(<level>) . . . . . . .  code assignment to several lists
 *F  CodeAsssListLevel(<level>)  . . code multiple assignment to several lists
 */
-extern  void            libGAP_CodeAssList ( void );
+extern  void            libGAP_CodeAssList ( libGAP_Int narg );
 
 extern  void            libGAP_CodeAsssList ( void );
 
-extern  void            libGAP_CodeAssListLevel (
+extern  void            libGAP_CodeAssListLevel ( libGAP_Int narg,
             libGAP_UInt                level );
 
 extern  void            libGAP_CodeAsssListLevel (
             libGAP_UInt                level );
 
-extern  void            libGAP_CodeUnbList ( void );
+extern  void            libGAP_CodeUnbList ( libGAP_Int narg );
 
 
 /****************************************************************************
@@ -1100,17 +1180,18 @@ extern  void            libGAP_CodeUnbList ( void );
 *F  CodeElmListLevel(<level>) . . . . . . . . code selection of several lists
 *F  CodeElmsListLevel(<level>)  . .  code multiple selection of several lists
 */
-extern  void            libGAP_CodeElmList ( void );
+extern  void            libGAP_CodeElmList ( libGAP_Int narg );
 
 extern  void            libGAP_CodeElmsList ( void );
 
 extern  void            libGAP_CodeElmListLevel (
-            libGAP_UInt                level );
+					  libGAP_Int narg,
+					  libGAP_UInt level);
 
 extern  void            libGAP_CodeElmsListLevel (
             libGAP_UInt                level );
 
-extern  void            libGAP_CodeIsbList ( void );
+extern  void            libGAP_CodeIsbList ( libGAP_Int narg );
 
 
 /****************************************************************************
